"""States module for managing EC2 VPC Endpoint Service Configurations."""
import copy
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    acceptance_required: bool = None,
    private_dns_name: str = None,
    network_load_balancer_arns: List[str] = None,
    gateway_load_balancer_arns: List[str] = None,
    supported_ip_address_types: List[str] = None,
    client_token: str = None,
    tags: Dict[str, str] or List = None,
) -> Dict[str, Any]:
    """
    Creates a VPC endpoint service to which service consumers (Amazon Web Services accounts, users, and IAM roles)
    can connect.

    Before you create an endpoint service, you must create one of the following for your service:

        * A Network Load Balancer. Service consumers connect to your service using an interface endpoint.

        * A Gateway Load Balancer. Service consumers connect to your service using a Gateway Load Balancer endpoint.

    If you set the private DNS name, you must prove that you own the private DNS domain name.
    For more information, see the Amazon Web Services PrivateLink Guide.

    Args:
        name(str, Optional): Idem name of the resource. Defaults to None.

        acceptance_required(bool, Optional): Indicates whether requests from service consumers to create an endpoint to your service must be
            accepted manually. Defaults to None.

        private_dns_name(str, Optional): (Interface endpoint configuration) The private DNS name to assign to the VPC endpoint service. Defaults to None.

        network_load_balancer_arns(List[str], Optional): The Amazon Resource Names (ARNs) of the Network Load Balancers. Defaults to None.

        gateway_load_balancer_arns(List[str], Optional): The Amazon Resource Names (ARNs) of the Gateway Load Balancers. Defaults to None.

        supported_ip_address_types(List[str], Optional): The supported IP address types. The possible values are ipv4 and ipv6. Defaults to None.

        client_token(str, Optional): Unique, case-sensitive identifier that you provide to ensure the idempotency of the request. For
            more information, see How to ensure idempotency. Defaults to None.

        tags (Dict or List, Optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of [{"Key": tag-key, "Value": tag-value}] to associate with the VPC. Each tag consists of a key name and an associated value. Defaults to None.

            * Key (str, Optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws:.
            * Value(str, Optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256 Unicode characters.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

          my-vpc-endpoint-service-configuration:
            aws.ec2.vpc_endpoint_service_configuration.present:
              - acceptance_required: bool
              - private_dns_name: string
              - network_load_balancer_arns:
                - value
              - gateway_load_balancer_arns:
                - value
              - supported_ip_address_types:
                - value
              - client_token: string
              - tags:
                  - Key: string
                    Value: string
    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    current_state = None

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        before = await hub.exec.aws.ec2.vpc_endpoint_service_configuration.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = current_state = copy.deepcopy(before["ret"])
        result["comment"].append(
            f"'aws.ec2.vpc_endpoint_service_configuration: {name}' already exists"
        )

    if current_state:
        # If there are changes in desired state from existing state
        changes = differ.deep_diff(
            current_state if current_state else {}, desired_state
        )

        if bool(changes.get("new")):
            if ctx.test:
                result["new_state"] = hub.tool.test_state_utils.generate_test_state(
                    enforced_state={}, desired_state=desired_state
                )
                result["comment"].append(
                    f"Would update aws.ec2.vpc_endpoint_service_configuration: '{name}'",
                )
                return result
            else:
                # Get new desired state for the vpc_endpoint_service_configuration
                desired_state = hub.tool.aws.ec2.vpc_endpoint_service_configuration.evaluate_update_desired_state(
                    ctx=ctx, current_state=current_state, desired_state=desired_state
                )

                # Update the resource
                update_ret = (
                    await hub.exec.aws.ec2.vpc_endpoint_service_configuration.update(
                        ctx,
                        **desired_state,
                    )
                )

                bool(update_ret["ret"])
                result["result"] = update_ret["result"]

                if result["result"]:
                    result["comment"].append(
                        f"Updated aws.ec2.vpc_endpoint_service_configuration '{name}'"
                    )
                else:
                    result["comment"].append(update_ret["comment"])
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            result["comment"].append(
                f"Would create aws.ec2.vpc_endpoint_service_configuration '{name}'",
            )
            return result

        create_ret = await hub.exec.aws.ec2.vpc_endpoint_service_configuration.create(
            ctx,
            **desired_state,
        )
        result["result"] = create_ret["result"]

        if result["result"]:
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.ec2.vpc_endpoint_service_configuration",
                name=name,
            )
            resource_id = create_ret["ret"].get("resource_id")
            # Safeguard for any future errors so that the resource_id is saved in the ESM
            result["new_state"] = dict(name=name, resource_id=resource_id)
        else:
            result["comment"].append(create_ret["comment"])

    if not result["result"]:
        # If there is any failure in create/update, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = dict(name=name, resource_id=resource_id)

    after = await hub.exec.aws.ec2.vpc_endpoint_service_configuration.get(
        ctx, name=name, resource_id=resource_id
    )
    result["new_state"] = after.ret
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Deletes the specified VPC endpoint service configurations. Before you can delete an endpoint service
    configuration, you must reject any Available or PendingAcceptance interface endpoint connections that are
    attached to the service.

    Args:
        name(str): Idem name of the resource.

        resource_id(str, Optional): The ID of the service. Defaults to None.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

            my-vpc-endpoint-service-configuration:
                aws.ec2.vpc_endpoint_service_configuration.absent:
                  - name: my-vpc-endpoint-service-configuration
                  - resource_id: value
    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # This is to make absent idempotent. If absent is run again, it would be a no-op
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.vpc_endpoint_service_configuration", name=name
        )
        return result

    before = await hub.exec.aws.ec2.vpc_endpoint_service_configuration.get(
        ctx, name=name, resource_id=resource_id
    )

    # Case: Error
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    # Case: Not Found
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.vpc_endpoint_service_configuration", name=name
        )
        return result

    result["old_state"] = before["ret"]

    if ctx.get("test", False):
        result["comment"].append(
            f"Would delete aws.ec2.vpc_endpoint_service_configuration '{name}'",
        )
        return result

    delete_ret = await hub.exec.aws.ec2.vpc_endpoint_service_configuration.delete(
        ctx,
        name=name,
        resource_id=resource_id,
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        # If there is any failure in delete, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = resource_id
        result["comment"].append(delete_ret["result"])

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.ec2.vpc_endpoint_service_configuration", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describes the VPC endpoint service configurations in your account (your services).

    Returns:
        Dict[str, Any]

    Example:

        .. code-block:: bash

            $ idem describe aws.ec2.vpc_endpoint_service_configuration
    """

    result = {}

    ret = await hub.exec.aws.ec2.vpc_endpoint_service_configuration.list(ctx)

    if not ret or not ret["result"]:
        hub.log.debug(
            f"Could not describe aws.ec2.vpc_endpoint_service_configuration {ret['comment']}"
        )
        return result

    for resource in ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "aws.ec2.vpc_endpoint_service_configuration.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""State module for AWS Neptune DB Parameter group."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

RESOURCE_TYPE = "aws.neptune.db_parameter_group"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    db_parameter_group_family: str,
    description: str,
    parameters: List[
        make_dataclass(
            """DB Parameter group configuration.""" "Parameter",
            [
                ("ParameterName", str),
                ("ParameterValue", str),
                ("ApplyMethod", str, field(default=None)),
                ("Description", str, field(default=None)),
                ("Source", str, field(default=None)),
                ("ApplyType", str, field(default=None)),
                ("DataType", str, field(default=None)),
                ("AllowedValues", str, field(default=None)),
                ("MinimumEngineVersion", str, field(default=None)),
                ("SupportedEngineModes", List[str], field(default=None)),
            ],
        )
    ] = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """Creates a new DB parameter group.

    A DB parameter group is initially created with the default parameters for the
    database engine used by the DB instance. To provide custom values for any of the parameters, you must modify the
    group after creating it using ModifyDBParameterGroup. Once you've created a DB parameter group, you need to
    associate it with your DB instance using ModifyDBInstance. When you associate a new DB parameter group with a
    running DB instance, you need to reboot the DB instance without failover for the new DB parameter group and
    associated settings to take effect.  After you create a DB parameter group, you should wait at least 5 minutes
    before creating your first DB instance that uses that DB parameter group as the default parameter group. This
    allows Amazon Neptune to fully complete the create action before the parameter group is used as the default for
    a new DB instance. This is especially important for parameters that are critical when creating the default
    database for a DB instance, such as the character set for the default database defined by the
    character_set_database parameter. You can use the Parameter Groups option of the Amazon Neptune console or the
    DescribeDBParameters command to verify that your DB parameter group has been created or modified.

    Args:
        name(str):
            An idem name of the resource.
        db_parameter_group_family(str):
            The DB parameter group family name. A DB parameter group can be associated with one and only one
            DB parameter group family, and can be applied only to a DB instance running a database engine
            and engine version compatible with that DB parameter group family.
        parameters(list, Optional):
            An array of parameter names, values, and the apply method for the parameter update.
            At least one parameter name, value, and apply method must be supplied; subsequent arguments are optional. Defaults to None.
        resource_id(str, Optional):
            AWS DBParameterGroup Name. Defaults to None.
        description(str):
            The description for the DB parameter group.
        tags(dict or list, Optional):
            Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the DB parameter group.
            Each tag consists of a key name and an associated value. Defaults to None.

            * Key (str, Optional):
                The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.

            * Value(str, Optional):
                The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_neptune_db_parameter_group]:
                aws.neptune.db_parameter_group.present:
                    - name: 'string'
                    - db_parameter_group_family: 'string'
                    - parameters:
                        - AllowedValues: 'string'
                          ApplyMethod: 'string'
                          ApplyType: 'string'
                          DataType: 'string'
                          Description: 'string'
                          IsModifiable: 'string'
                          MinimumEngineVersion: 'string'
                          ParameterName: 'string'
                          ParameterValue: 'string'
                          Source: 'string'
                    - resource_id: 'string'
                    - description: 'string'
                    - tags:
                        - Key: 'string'
                          Value: 'string'
                        - Key: 'string'
                          Value: 'string'

    Returns:
        Dict[str, Dict[str,Any]]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.neptune.db_parameter_group.present:
                - name: value
                - db_parameter_group_name: value
                - db_parameter_group_family: value
                - description: value
                - parameters:
                  - AllowedValues: 10-2147483647
                    ApplyMethod: pending-reboot
                    ApplyType: static
                    DataType: integer
                    Description: Graph query timeout (ms).
                    IsModifiable: true
                    MinimumEngineVersion: 1.0.1.0
                    ParameterName: neptune_query_timeout
                    ParameterValue: '120001'
                    Source: user
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    desired_state = {
        "name": name,
        "resource_id": name,
        "db_parameter_group_family": db_parameter_group_family,
        "description": description,
        "parameters": parameters,
        "tags": tags,
    }
    # set a flag to keep track if the resource is newly created
    is_created = False

    if not resource_id:
        # if there is no resource_id, it is expected that we create the resource
        # if this is test run, just generate the new state and set in result
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            resource_id = desired_state["resource_id"]
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        else:
            # if this is not a test run, create the real resource
            create_db_parameter_group_result = (
                await hub.exec.boto3.client.neptune.create_db_parameter_group(
                    ctx,
                    DBParameterGroupName=name,
                    DBParameterGroupFamily=db_parameter_group_family,
                    Description=description,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                )
            )
            # if create failed, fail the request
            if not create_db_parameter_group_result["result"]:
                result["comment"] = create_db_parameter_group_result["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
            resource_id = create_db_parameter_group_result["ret"]["DBParameterGroup"][
                "DBParameterGroupName"
            ]
        is_created = True

    # The argument "parameters" can only be modified as per AWS docs, so if user specifies
    # this during create, idem will first create the db_parameter_group and then modify its parameters
    # P.S. There is no way to create these parameters as per AWS doc.
    update_db_parameter_group_result = (
        await hub.tool.aws.neptune.db_parameter_group.update(
            ctx,
            name=name,
            resource_id=resource_id,
            desired_state=desired_state,
            is_created=is_created,
        )
    )
    if not (
        update_db_parameter_group_result
        and update_db_parameter_group_result["result"]
        and update_db_parameter_group_result["new_state"]
    ):
        result["result"] = False
        result["comment"] = update_db_parameter_group_result["comment"]
        return result

    result["old_state"] = update_db_parameter_group_result["old_state"]
    result["new_state"] = update_db_parameter_group_result["new_state"]
    # update the result comment if the resource is not being created
    if not is_created:
        result["comment"] = update_db_parameter_group_result["comment"]

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes a specified DBParameterGroup.

    The DBParameterGroup to be deleted can't be associated with any DB
    instances.

    Args:
        name(str):
            An idem name of the resource.
        resource_id(str, Optional):
            AWS Neptune DB Parameter Group Name. Defaults to None
            Idem automatically considers this resource as absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_neptune_db_parameter_group]:
                aws.neptune.db_parameter_group.absent:
                    - name: 'string'
                    - resource_id: 'string'

    Returns:
        Dict[str, Dict[str,Any]]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.neptune.db_parameter_group.absent:
                - name: value
                - resource_id: value
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type=RESOURCE_TYPE, name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.aws.neptune.db_parameter_group.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        return result
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.neptune.delete_db_parameter_group(
            ctx,
            DBParameterGroupName=resource_id,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns a list of DBParameterGroup descriptions.

    Returns:
        Dict[str, Dict[str,Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.neptune.db_parameter_group
    """
    result = {}
    ret = await hub.exec.aws.neptune.db_parameter_group.list(
        ctx, name="aws.neptune.db_parameter_group.describe"
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe {RESOURCE_TYPE} {ret['comment']}")
        return result

    for resource in ret["ret"]:
        result[resource.get("resource_id")] = {
            "aws.neptune.db_parameter_group.present": [resource]
        }
    return result



from typing import Any
from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: List[Dict[str, Any]],
    new_value: List[Dict[str, Any]],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in its "present" state

    Args:
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process

    Example:
        - security_group_ids:
          - sg-0839b8c0df83f79f4
          - sg-0c8313e7fd3d8e48b
    """
    if old_value == new_value:
        return True

    ret = await hub.exec.boto3.client.ec2.modify_instance_attribute(
        ctx,
        InstanceId=resource.id,
        Groups=new_value,
    )
    if not ret.result:
        comments.append(ret.comment)

    return ret.result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_integration_response_to_present(
    hub,
    raw_resource: Dict[str, Any],
    resource_id: str,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "responseParameters": "response_parameters",
            "responseTemplates": "response_templates",
            "contentHandling": "content_handling",
            "selectionPattern": "selection_pattern",
        }
    )

    resource_translated = {
        "resource_id": resource_id,
        "rest_api_id": resource_id.split("-")[0],
        "parent_resource_id": resource_id.split("-")[1],
        "http_method": resource_id.split("-")[2],
        "status_code": resource_id.split("-")[3],
    }

    if idem_resource_name:
        resource_translated["name"] = idem_resource_name

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    old_state: Dict[str, Any],
    updatable_parameters: Dict[str, Any],
):
    """

    Updates an API Gateway Integration Response.

    Args:
        hub:
        ctx:
        old_state(dict):
            Previous state of the resource.

        updatable_parameters(dict):
            Parameters from SLS File. Add is not supported for content_handling and selection_pattern.

    Returns:
        {"result": True|False, "comment": A message List, "ret": None|Dict}

    """

    result = dict(comment=(), result=True, ret=None)
    patch_ops = []

    new_state = copy.deepcopy(old_state)
    updated = False
    key_final = ""
    path_final = ""
    items_final = ""

    for key, value in updatable_parameters.items():
        if key == "content_handling":
            key_final = "contentHandling"
        elif key == "response_parameters":
            key_final = "responseParameters"
        elif key == "response_templates":
            key_final = "responseTemplates"
        elif key == "selection_pattern":
            key_final = "selectionPattern"

        if (
            not isinstance(value, str)
            and value is not None
            and value.items() is not None
        ):
            for items_key, items_val in value.items():
                path_final = (f"/{key_final}/{items_key}",)
                items_final = items_val
        else:
            path_final = f"/{key_final}"
            items_final = value

        if old_state.get(key) and value is not None and value != old_state.get(key):
            patch_ops.append(
                {
                    "op": "replace",
                    "path": path_final,
                    "value": str(items_final)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            updated = True
        elif old_state.get(key) is None and value is not None:
            patch_ops.append(
                {
                    "op": "add",
                    "path": path_final,
                    "value": str(items_final)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            new_state[key] = value
            updated = True

    if ctx.get("test", False):
        if updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.integration_response",
                name=old_state["name"],
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.integration_response",
                name=old_state["name"],
            )
        return result

    if updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_integration_response(
            ctx,
            restApiId=old_state["rest_api_id"],
            resourceId=old_state["parent_resource_id"],
            httpMethod=old_state["http_method"],
            statusCode=old_state["status_code"],
            patchOperations=patch_ops,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.integration_response",
                name=old_state["name"],
            )
            result["ret"] = new_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.integration_response", name=old_state["name"]
        )

    return result



"""Exec module for managing budgets."""
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, resource_id: str) -> Dict:
    """Retrieves the budget details.

    Args:
        name(str): An Idem name of the AWS Budget.
        resource_id(str): Resource ID to identify the AWS Budget.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS budget in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.budgets.budget.get name="idem_name" resource_id="budget_name"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.budgets.budget.get(
                    ctx, name=name, resource_id=resource_id
                )
    """
    result = dict(comment=[], ret=None, result=True)
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]
    ret = await hub.exec.boto3.client.budgets.describe_budget(
        ctx, AccountId=account_id, BudgetName=resource_id
    )
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.budgets.budget", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result[
        "ret"
    ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present_async(
        ctx,
        account_id=account_id,
        raw_resource=ret["ret"].get("Budget"),
        idem_resource_name=name,
    )
    return result


async def list_(
    hub,
    ctx,
) -> Dict:
    """Fetch a list of budgets from AWS.

    The function returns empty list when no resource is found.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS budget list in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.budgets.budget.list

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.budgets.budget.list(ctx)

    """
    result = dict(comment=[], ret=[], result=True)
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]
    ret = await hub.exec.boto3.client.budgets.describe_budgets(
        ctx, AccountId=account_id
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Budgets"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.budgets.budget", name="Budgets"
            )
        )
        return result
    for budget in ret["ret"]["Budgets"]:
        result["ret"].append(
            await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present_async(
                ctx,
                account_id=account_id,
                raw_resource=budget,
                idem_resource_name=budget.get("BudgetName"),
            )
        )
    return result



NAME_PARAMETER = {
    "default": None,
    "doc": "An Idem name of the resource",
    "param_type": "str",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

RESOURCE_ID_PARAMETER = {
    "default": None,
    "doc": "An identifier of the resource in the provider",
    "param_type": "str",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

OLD_TAGS_PARAMETER = {
    "default": None,
    "doc": "Dict in the format of {tag-key: tag-value}",
    "param_type": "dict",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

NEW_TAGS_PARAMETER = {
    "default": None,
    "doc": "Dict in the format of {tag-key: tag-value}",
    "param_type": "dict",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

TAGS_PARAMETER = {
    "default": None,
    "doc": """Dict in the format of {tag-key: tag-value} or List of tags in the format of [{"Key": tag-key, "Value": tag-value}] to associate with the VPC. Each tag consists of a key name and an associated value. Defaults to None.

            * Key (str, Optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws:.
            * Value(str, Optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256 Unicode characters.""",
    "param_type": "Dict[str, Any] or List",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

RAW_RESOURCE_PARAMETER = {
    "default": None,
    "doc": "The raw representation of the resource in the provider",
    "param_type": "dict",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}



"""RDS Tag operations"""
import copy
from typing import Any
from typing import Dict


async def update_rds_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """
    Update tags of AWS RDS resources

    Args:
        resource_arn (str):
            AWS resource arn

        old_tags (dict):
            Dict in the format of {tag-key: tag-value}

        new_tags (dict):
            Dict in the format of {tag-key: tag-value}

    Returns:
        {"result": True|False, "comment": "A message", "ret": dict of updated tags}

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.rds.tag.update_rds_tags resource_arn="some_arn" old_tags='{"old_tag": "old_val"}' new_tags='{"new_tag": "new_val"}'

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.rds.tag.update_rds_tags
                - kwargs:
                    resource_arn: some_arn
                    old_tags:
                      old_tag: old_val
                    new_tags:
                      new_tag: new_val
    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.rds.remove_tags_from_resource(
                ctx,
                ResourceName=resource_arn,
                TagKeys=list(tags_to_remove),
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.rds.add_tags_to_resource(
                ctx,
                ResourceName=resource_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    result["comment"] = result["comment"] + (
        f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",
    )
    return result



"""States module for managing Rds Db Proxy Endpoints."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_proxy_name: str,
    db_proxy_endpoint_name: str,
    vpc_subnet_ids: List[str],
    vpc_security_group_ids: List[str] = None,
    target_role: str = None,
    tags: Dict[str, str] = None,
    resource_id: str = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Creates a DBProxyEndpoint. Only applies to proxies that are associated with Aurora DB clusters. You can use DB
    proxy endpoints to specify read/write or read-only access to the DB cluster. You can also use DB proxy endpoints
    to access a DB proxy through a different VPC than the proxy's default VPC.

    Args:
        db_proxy_name(str): The name of the DB proxy associated with the DB proxy endpoint that you create.

        db_proxy_endpoint_name(str): The name of the DB proxy endpoint to create.

        vpc_subnet_ids(List[str]): The VPC subnet IDs for the DB proxy endpoint that you create. You can specify a different set of
            subnet IDs than for the original DB proxy.

        name(str): Idem name of the resource.

        vpc_security_group_ids(List[str], Optional): The VPC security group IDs for the DB proxy endpoint that you create. You can specify a
            different set of security group IDs than for the original DB proxy. The default is the default
            security group for the VPC. Defaults to None.

        target_role(str, Optional): A value that indicates whether the DB proxy endpoint can be used for read/write or read-only
            operations. The default is READ_WRITE. The only role that proxies for RDS for Microsoft SQL
            Server support is READ_WRITE. Defaults to None.

        tags(Dict[str, str], Optional): The tags to apply to the resource.

        resource_id(str, Optional): Db_proxy_endpoint unique ID. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Cluster.

            * create (*dict, Optional*):
                Timeout configuration for creating DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

            * update(*dict, Optional*):
                Timeout configuration for updating DB Cluster

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls


          idem_test_aws_auto.rds.db_proxy_endpoint_is_present:
              aws_auto.aws_auto.rds.db_proxy_endpoint.present:
              - db_proxy_name: string
              - db_proxy_endpoint_name: string
              - vpc_subnet_ids:
                - value
              - vpc_security_group_ids:
                - value
              - target_role: string
              - tags:
                  key: value


    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    current_state = None

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        before = await hub.exec.aws.rds.db_proxy_endpoint.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = current_state = copy.deepcopy(before["ret"])
        result["comment"] += hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.rds.db_proxy_endpoint", name=name
        )

    if current_state:
        # If there are changes in desired state from existing state
        changes = differ.deep_diff(
            current_state if current_state else {}, desired_state
        )
        if bool(changes.get("new")):
            if ctx.test:
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={}, desired_state=desired_state
                )
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.rds.db_proxy_endpoint", name=name
                )
                return result
            else:
                # Update the resource
                update_ret = await hub.exec.aws.rds.db_proxy_endpoint.update(
                    ctx,
                    resource_id=resource_id,
                    tags=tags,
                    new_db_proxy_endpoint_name=db_proxy_endpoint_name,
                    vpc_security_group_ids=vpc_security_group_ids,
                    name=name,
                )
                result["result"] = update_ret["result"]
                if result["result"]:
                    resource_id = update_ret["ret"].get("resource_id", resource_id)
                    result["comment"] += hub.tool.aws.comment_utils.update_comment(
                        resource_type="aws.rds.db_proxy_endpoint", name=name
                    )
                else:
                    result["comment"].append(update_ret["comment"])
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            result["comment"] += hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.rds.db_proxy_endpoint", name=name
            )
            return result

        create_ret = await hub.exec.aws.rds.db_proxy_endpoint.create(
            ctx,
            **desired_state,
        )
        result["result"] = create_ret["result"]

        if result["result"]:
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.rds.db_proxy_endpoint", name=name
            )
            resource_id = create_ret["ret"].get("resource_id")
            # Safeguard for any future errors so that the resource_id is saved in the ESM
            result["new_state"] = dict(name=name, resource_id=resource_id)
        else:
            result["comment"].append(create_ret["comment"])

    if not result["result"]:
        # If there is any failure in create/update, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = dict(name=name, resource_id=resource_id)

    after = await hub.exec.aws.rds.db_proxy_endpoint.get(
        ctx, name=name, resource_id=resource_id
    )
    result["new_state"] = after.ret
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: make_dataclass(
        """Specifies timeout for deletion of DB Proxy Endpoint.""" "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=40)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """

    Deletes a DBProxyEndpoint. Doing so removes the ability to access the DB proxy using the endpoint that you
    defined. The endpoint that you delete might have provided capabilities such as read/write or read-only
    operations, or using a different VPC than the DB proxy's default VPC.

    Args:
        name(str): Idem name of the resource.

        resource_id(str): The name of the DB proxy endpoint to delete.

        timeout(dict, Optional):
            Timeout configuration for delete of AWS RDS DB Proxy.

            * delete (*dict, Optional*):
                Timeout configuration for deleting DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls


            idem_test_aws_auto.rds.db_proxy_endpoint_is_absent:
              aws_auto.aws_auto.rds.db_proxy_endpoint.absent:
              - db_proxy_endpoint_name: string


    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    if not resource_id:
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_proxy_endpoint", name=name
        )
        return result

    before = await hub.exec.aws.rds.db_proxy_endpoint.get(
        ctx, name=name, resource_id=resource_id
    )

    # Case: Error
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    # Case: Not Found
    if not before["ret"]:
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_proxy_endpoint", name=name
        )
        return result

    result["old_state"] = before["ret"]

    if ctx.get("test", False):
        result["comment"].append(
            f"Would delete aws.rds.db_proxy_endpoint '{name}'",
        )
        return result

    delete_ret = await hub.exec.aws.rds.db_proxy_endpoint.delete(
        ctx, name=name, resource_id=resource_id, timeout=timeout
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        # If there is any failure in delete, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = resource_id
        result["comment"].append(delete_ret["comment"])
    else:
        result["comment"] += hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.rds.db_proxy_endpoint", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns information about DB proxy endpoints.


    Returns:
        Dict[str, Any]

    Example:

        .. code-block:: bash

            $ idem describe aws_auto.rds.db_proxy_endpoint
    """

    result = {}
    ret = await hub.exec.aws.rds.db_proxy_endpoint.list(ctx)

    if not ret or not ret["result"]:
        hub.log.debug(f"Could not describe aws.rds.db_proxy_endpoint {ret['comment']}")
        return result

    for resource in ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "aws.rds.db_proxy_endpoint.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""Exec module for managing Amazon ECR repository policies."""
from typing import Any
from typing import Dict


async def get(
    hub, ctx, name: str, resource_id: str, registry_id: str = None
) -> Dict[str, Any]:
    """Retrieves the repository policy for the specified repository.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            The name of the ECR repository in Amazon Web Services.
        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository
            to search. If you do not specify a registry, the default registry is assumed.

    Returns:
        Dict[str, Any]:
            Return the repository policy for the specified repository.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.ecr.repository_policy.get name='idem_name' resource_id='ecr_repository_name'

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ecr.repository_policy.get
                - kwargs:
                    name: 'idem_name'
                    resource_id: 'ecr_repository_name'
    """
    result = dict(comment=[], ret=None, result=True)

    get_policy_ret = await hub.exec.boto3.client.ecr.get_repository_policy(
        ctx=ctx, registryId=registry_id, repositoryName=resource_id
    )

    # Case: Error
    if not get_policy_ret["result"]:
        # Do not return success=false when it is not found.
        if "RepositoryNotFoundException" in str(
            get_policy_ret["comment"]
        ) or "RepositoryPolicyNotFoundException" in str(get_policy_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ecr.repository_policy", name=name
                )
            )
            result["comment"] += list(get_policy_ret["comment"])
            return result

        result["comment"] += list(get_policy_ret["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    if not get_policy_ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ecr.repository_policy", name=resource_id
            )
        )
        return result

    # Case: Result found
    result[
        "ret"
    ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_policy_to_present(
        raw_resource=get_policy_ret.get("ret"),
        idem_resource_name=name,
    )

    return result



"""List of possible functions for known operations."""
import re

DESCRIBE_FUNCTIONS = ["get", "head", "search", "describe"]

LIST_FUNCTIONS = ["list", "describe", "search", "get"]

DELETE_FUNCTIONS = [
    "delete",
    "disassociate",
    "reject",
    "deallocate",
    "unassign",
    "deregister",
    "deprovision",
    "revoke",
    "release",
    "terminate",
    "cancel",
    "disable",
]

CREATE_FUNCTIONS = [
    "create",
    "associate",
    "accept",
    "allocate",
    "assign",
    "register",
    "provision",
    "authorize",
    "run",
    "enable",
    "put",
    "publish",
    "request",
    "put",
    "add",
    "apply",
]

UPDATE_FUNCTIONS = ["modify", "update", "put", "apply"]


def get_possible_func_type(hub, resource_name: str, operation: str):
    if re.match(
        _get_regex_pattern(DESCRIBE_FUNCTIONS), operation
    ) and operation.endswith(resource_name):
        return "get"

    if (
        re.match(_get_regex_pattern(LIST_FUNCTIONS), operation)
        and (
            operation.endswith(f"{resource_name}s")
            or operation.endswith(f"{resource_name}es")
        )
    ) or re.match("(describe).*", operation):
        return "list"

    if re.match(_get_regex_pattern(CREATE_FUNCTIONS), operation):
        return "create"

    if re.match(_get_regex_pattern(UPDATE_FUNCTIONS), operation):
        return "update"

    if re.match(_get_regex_pattern(DELETE_FUNCTIONS), operation):
        return "delete"

    return "tools"


def _get_regex_pattern(possible_function):
    regex_str = "|".join(possible_function)
    regex_str = f"({regex_str}).*"
    return regex_str



"""Exec module for managing Rds Db Proxy Endpoints."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["soft_fail"]

__func_alias__ = {"list_": "list"}

create_waiter_acceptors = [
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "success",
        "argument": "DBProxyEndpoints[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "creating",
        "state": "retry",
        "argument": "DBProxyEndpoints[].Status",
    },
]

update_waiter_acceptors = [
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "success",
        "argument": "DBProxyEndpoints[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "modifying",
        "state": "retry",
        "argument": "DBProxyEndpoints[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "resetting-master-credentials",
        "state": "retry",
        "argument": "DBProxyEndpoints[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "renaming",
        "state": "retry",
        "argument": "DBProxyEndpoints[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "upgrading",
        "state": "retry",
        "argument": "DBProxyEndpoints[].Status",
    },
]
delete_waiter_acceptors = [
    {
        "matcher": "error",
        "expected": "DBProxyEndpointNotFoundFault",
        "state": "success",
        "argument": "Error.Code",
    },
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "retry",
        "argument": "DBProxyEndpoints[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "deleting",
        "state": "retry",
        "argument": "DBProxyEndpoints[].Status",
    },
]


async def get(
    hub,
    ctx,
    resource_id: str = None,
    db_proxy_name: str = None,
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
    name: str = None,
) -> Dict[str, Any]:
    """

    Returns information about DB proxy endpoints.

    Args:
        resource_id(str, Optional): The name of a DB proxy endpoint to describe. If you omit this parameter, the output includes
            information about all DB proxy endpoints associated with the specified proxy. Defaults to None.

        db_proxy_name(str, Optional): The name of the DB proxy whose endpoints you want to describe. If you omit this parameter, the
            output includes information about all DB proxy endpoints associated with all your DB proxies. Defaults to None.


        filters(List[dict[str, Any]], Optional): This parameter is not currently supported. Defaults to None.

            * Name (str): The name of the filter. Filter names are case-sensitive.

            * Values (List[str]): One or more filter values. Filter values are case-sensitive.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            unmanaged_resource:
              exec.run:
                - path: aws.rds.db_proxy_endpoint.get
                - kwargs:
                  resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_endpoint.get resource_id=value
    """

    result = dict(comment=[], ret=None, result=True)

    get_db_proxy_endpoint = await hub.exec.boto3.client.rds.describe_db_proxy_endpoints(
        ctx=ctx,
        **{
            "DBProxyName": db_proxy_name,
            "DBProxyEndpointName": resource_id,
            "Filters": filters,
        },
    )

    # Case: Error
    if not get_db_proxy_endpoint["result"]:
        # Do not return success=false when it is not found.
        # Most of the resources would return "*NotFound*" type of exception when it is 404
        if "DBProxyEndpointNotFoundFault" in str(
            get_db_proxy_endpoint["comment"]
        ) or "DBProxyNotFoundFault" in str(get_db_proxy_endpoint["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.rds.db_proxy_endpoint", name=resource_id
                )
            )
            result["comment"].append(get_db_proxy_endpoint["comment"])
            return result

        result["comment"].append(get_db_proxy_endpoint["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    if not get_db_proxy_endpoint["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_proxy_endpoint", name=resource_id
            )
        )
        return result

    if len(get_db_proxy_endpoint["ret"]["DBProxyEndpoints"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_id=resource_id, resource_type="aws.rds.db_proxy_endpoint"
            )
        )

    raw_resource = get_db_proxy_endpoint["ret"]["DBProxyEndpoints"][0]

    result[
        "ret"
    ] = await hub.tool.aws.rds.db_proxy_endpoint.convert_raw_resource_to_present_async(
        ctx=ctx,
        resource_id=resource_id
        if resource_id
        else raw_resource.get("DBProxyEndpointName"),
        raw_resource=raw_resource,
        idem_resource_name=name,
    )

    return result


async def list_(
    hub,
    ctx,
    db_proxy_name: str = None,
    resource_id: str = None,
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Returns information about DB proxy endpoints.

    Args:
        db_proxy_name(str, Optional): The name of the DB proxy whose endpoints you want to describe. If you omit this parameter, the
            output includes information about all DB proxy endpoints associated with all your DB proxies. Defaults to None.

        resource_id(str, Optional): The name of a DB proxy endpoint to describe. If you omit this parameter, the output includes
            information about all DB proxy endpoints associated with the specified proxy. Defaults to None.

        filters(List[dict[str, Any]], Optional): This parameter is not currently supported. Defaults to None.

            * Name (str): The name of the filter. Filter names are case-sensitive.

            * Values (List[str]): One or more filter values. Filter values are case-sensitive.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        Resource State:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.rds.db_proxy_endpoint.list
                - kwargs:


        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_endpoint.list

        Describe call from the CLI:

        .. code-block:: bash

            $ idem describe aws.rds.db_proxy_endpoint

    """

    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.rds.describe_db_proxy_endpoints(
        ctx=ctx,
        **{
            "DBProxyName": db_proxy_name,
            "DBProxyEndpointName": resource_id,
            "Filters": filters,
            "name": name,
        },
    )

    if not ret["result"]:
        result["comment"].append(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"].get("DBProxyEndpoints"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.rds.db_proxy_endpoint", name=None
            )
        )
        return result

    for resource in ret["ret"]["DBProxyEndpoints"]:
        result["ret"].append(
            await hub.tool.aws.rds.db_proxy_endpoint.convert_raw_resource_to_present_async(
                ctx=ctx,
                resource_id=resource.get("DBProxyEndpointName"),
                raw_resource=resource,
                idem_resource_name=name
                if name
                else resource.get("DBProxyEndpointName"),
            )
        )
    return result


async def create(
    hub,
    ctx,
    db_proxy_name: str,
    db_proxy_endpoint_name: str,
    vpc_subnet_ids: List[str],
    vpc_security_group_ids: List[str] = None,
    target_role: str = None,
    tags: Dict[str, str] = None,
    resource_id: str = None,
    name: str = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Creates a DBProxyEndpoint. Only applies to proxies that are associated with Aurora DB clusters. You can use DB
    proxy endpoints to specify read/write or read-only access to the DB cluster. You can also use DB proxy endpoints
    to access a DB proxy through a different VPC than the proxy's default VPC.

    Args:
        db_proxy_name(str): The name of the DB proxy associated with the DB proxy endpoint that you create.

        db_proxy_endpoint_name(str): The name of the DB proxy endpoint to create.

        vpc_subnet_ids(List[str]): The VPC subnet IDs for the DB proxy endpoint that you create. You can specify a different set of
            subnet IDs than for the original DB proxy.

        vpc_security_group_ids(List[str], Optional): The VPC security group IDs for the DB proxy endpoint that you create. You can specify a
            different set of security group IDs than for the original DB proxy. The default is the default
            security group for the VPC. Defaults to None.

        target_role(str, Optional): A value that indicates whether the DB proxy endpoint can be used for read/write or read-only
            operations. The default is READ_WRITE. The only role that proxies for RDS for Microsoft SQL
            Server support is READ_WRITE. Defaults to None.

        tags(Dict[str, str], Optional): The tags to apply to the resource.

        resource_id(str, Optional): Db_proxy_endpoint unique ID. Defaults to None.

        name(str, Optional): Idem name of the resource. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Cluster.

            * create (*dict, Optional*):
                Timeout configuration for creating DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

            * update(*dict, Optional*):
                Timeout configuration for updating DB Cluster

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Examples:
        Using in a state:

        .. code-block:: sls

            resource_is_present:
              aws.rds.db_proxy_endpoint.present:
                - db_proxy_name: value
                - db_proxy_endpoint_name: value
                - vpc_subnet_ids: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_endpoint.create db_proxy_name=value, db_proxy_endpoint_name=value, vpc_subnet_ids=value
    """

    result = dict(comment=[], ret={}, result=True)

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    resource_to_raw_input_mapping = {
        "db_proxy_name": "DBProxyName",
        "db_proxy_endpoint_name": "DBProxyEndpointName",
        "vpc_subnet_ids": "VpcSubnetIds",
        "vpc_security_group_ids": "VpcSecurityGroupIds",
        "target_role": "TargetRole",
    }

    payload = {}
    for key, value in desired_state.items():
        if key in resource_to_raw_input_mapping.keys() and value is not None:
            payload[resource_to_raw_input_mapping[key]] = desired_state.get(key)
    if tags:
        payload["Tags"] = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)

    ret = await hub.exec.boto3.client.rds.create_db_proxy_endpoint(ctx, **payload)

    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] += ret["comment"]
        return result

    raw_resource = ret["ret"].get("DBProxyEndpoint", {})
    result["ret"]["resource_id"] = raw_resource.get("DBProxyEndpointName")
    result["ret"]["name"] = name

    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=20,
        default_max_attempts=40,
        timeout_config=timeout.get("create") if timeout else None,
    )
    cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
        name="DBProxyEndpointCreated",
        operation="DescribeDBProxyEndpoints",
        argument=["DBProxyEndpoints[].Status"],
        acceptors=create_waiter_acceptors,
        client=await hub.tool.boto3.client.get_client(ctx, "rds"),
    )
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "rds",
            "DBProxyEndpointCreated",
            cluster_waiter,
            DBProxyEndpointName=result["ret"]["resource_id"],
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] += (str(e),)
        result["result"] = False
        return result

    result["comment"] += hub.tool.aws.comment_utils.create_comment(
        resource_type="aws.rds.db_proxy_endpoint", name=name
    )

    return result


async def update(
    hub,
    ctx,
    resource_id: str,
    tags: Dict[str, str] = None,
    db_proxy_endpoint_name: str = None,
    new_db_proxy_endpoint_name: str = None,
    vpc_security_group_ids: List[str] = None,
    name: str = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Changes the settings for an existing DB proxy endpoint.

    Args:
        resource_id(str): The name of the DB proxy associated with the DB proxy endpoint that you want to modify.

        tags(dict, Optional): The tags to apply to the resource. Defaults to None.

        db_proxy_endpoint_name(str, Optional): Name of the DBProxyEndpoint.

        new_db_proxy_endpoint_name(str, Optional): The new identifier for the DBProxyEndpoint. An identifier must begin with a letter and must
            contain only ASCII letters, digits, and hyphens; it can't end with a hyphen or contain two
            consecutive hyphens. Defaults to None.

        vpc_security_group_ids(List[str], Optional): The VPC security group IDs for the DB proxy endpoint. When the DB proxy endpoint uses a
            different VPC than the original proxy, you also specify a different set of security group IDs
            than for the original proxy. Defaults to None.

        name(str, Optional): Idem name of the resource. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Cluster.

            * create (*dict, Optional*):
                Timeout configuration for creating DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

            * update(*dict, Optional*):
                Timeout configuration for updating DB Cluster

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Examples:
        Using in a state:

        .. code-block:: sls

            resource_is_present:
              aws.rds.db_proxy_endpoint.present:
                - db_proxy_endpoint_name: value
                - tags: value
                - resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_endpoint.update db_proxy_endpoint_name=value, tags=value, resource_id=value
    """

    result = dict(comment=[], ret={}, result=True)

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    payload = {}

    resource_to_raw_input_mapping = {
        "resource_id": "DBProxyEndpointName",
        "new_db_proxy_endpoint_name": "NewDBProxyEndpointName",
        "vpc_security_group_ids": "VpcSecurityGroupIds",
    }

    for key, value in desired_state.items():
        if key in resource_to_raw_input_mapping.keys() and value is not None:
            payload[resource_to_raw_input_mapping[key]] = desired_state.get(key)

    if payload:
        ret = await hub.exec.boto3.client.rds.modify_db_proxy_endpoint(ctx, **payload)
        if not ret["result"]:
            result["result"] = False
            result["comment"].append(
                f"Could not update aws.rds.db_proxy_endpoint '{name}'",
            )
            result["comment"].append(ret["comment"])
            return result

        result["comment"] += hub.tool.aws.comment_utils.update_comment(
            resource_type="aws.rds.db_proxy_endpoint", name=name
        )
        raw_resource = ret["ret"].get("DBProxyEndpoint", {})
        result["ret"]["resource_id"] = raw_resource.get("DBProxyEndpointName")
        get_tags_ret = await hub.tool.aws.rds.tag.get_tags_for_resource(
            ctx, resource_arn=raw_resource.get("DBProxyEndpointArn")
        )

        if get_tags_ret["result"]:
            current_tags = get_tags_ret.get("ret", {})
            update_tags_ret = await hub.tool.aws.rds.tag.update_tags(
                ctx,
                resource_arn=raw_resource.get("DBProxyEndpointArn"),
                old_tags=current_tags,
                new_tags=tags,
            )
            if not update_tags_ret["result"]:
                result["result"] = False
                result["comment"] += update_tags_ret["comment"]
                return result

        result["ret"]["resource_id"] = resource_id
        result["ret"]["name"] = name

        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=20,
            default_max_attempts=40,
            timeout_config=timeout.get("update") if timeout else None,
        )
        cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="DBProxyEndpointUpdated",
            operation="DescribeDBProxyEndpoints",
            argument=["DBProxyEndpoints[].Status"],
            acceptors=update_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "rds"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "rds",
                "DBProxyEndpointUpdated",
                cluster_waiter,
                DBProxyEndpointName=result["ret"]["resource_id"],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] += (str(e),)
            result["result"] = False
            return result

    return result


async def delete(
    hub,
    ctx,
    resource_id: str,
    name: str = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Deletes a DBProxyEndpoint. Doing so removes the ability to access the DB proxy using the endpoint that you
    defined. The endpoint that you delete might have provided capabilities such as read/write or read-only
    operations, or using a different VPC than the DB proxy's default VPC.

    Args:
        resource_id(str): The name of the DB proxy endpoint to delete.

        name(str, Optional): Idem name of the resource. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Proxy Endpoint.

            * delete (*dict, Optional*):
                Timeout configuration for deleting DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            resource_is_absent:
              aws.rds.db_proxy_endpoint.absent:
                - resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_endpoint.delete resource_id=value
    """

    result = dict(comment=[], ret=None, result=True)

    delete_ret = await hub.exec.boto3.client.rds.delete_db_proxy_endpoint(
        ctx,
        **{"DBProxyEndpointName": resource_id},
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        result["comment"] = delete_ret["comment"]
        result["result"] = False
        return result

    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=20,
        default_max_attempts=40,
        timeout_config=timeout.get("delete") if timeout else None,
    )
    cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
        name="DBProxyEndpointDeleted",
        operation="DescribeDBProxyEndpoints",
        argument=["DBProxyEndpoints[].Status"],
        acceptors=delete_waiter_acceptors,
        client=await hub.tool.boto3.client.get_client(ctx, "rds"),
    )
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "rds",
            "DBProxyEndpointDeleted",
            cluster_waiter,
            DBProxyEndpointName=resource_id,
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] += (str(e),)
        result["result"] = False
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.rds.db_proxy_endpoint", name=name
    )

    return result



"""State module for adding/deleting members from Guardduty."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": ["aws.guardduty.organization_admin_account.absent"],
    },
    "present": {
        "require": [
            "aws.guardduty.detector.present",
            "aws.guardduty.organization_admin_account.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    detector_id: str,
    account_id: str,
    email: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Add member to AWS Guardduty.

    Creates a member account under Guardduty by specifying Amazon Web Services account ID for the member account.
    This step is a prerequisite for managing the associated member accounts either by invitation or
    through an organization. When using Create Members as an organizations delegated administrator this action will
    enable GuardDuty in the added member accounts, with the exception of the organization delegated administrator
    account, which must enable GuardDuty prior to being added as a member. If you are adding accounts by invitation
    use this action after GuardDuty has been enabled in potential member accounts and before using  Invite Members .

    Args:
        name(str):
            An Idem name of the resource.

        detector_id(str):
            AWS Guardduty detector ID

        account_id(str):
            The member account ID.

        email(str):
            The email address of the member account.

        resource_id(str, Optional):
            An identifier refers to an existing resource. The format is <detector_id>:<account_id>

    Request Syntax:
      .. code-block:: sls
        [idem_test_aws_guardduty_member]:
          aws.guardduty.member.present:
            - name: 'string'
            - detector_id: 'string'
            - account_id: 'string'
            - email: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.guardduty.member.present:
                - name: 'create_members'
                - detector_id: "68c25425ab84ea0dcae26311eddacd34"
                - resource_id: "68c25425ab84ea0dcae26311eddacd34:496603212238"
                - account_id: "496603212238"
                - email: "xyz@vmware.com"
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    update_needed = False
    revised_account_details = []
    # pre-processing for create_members call
    entry = {"AccountId": account_id, "Email": email}
    revised_account_details.append(entry)
    before = await hub.exec.aws.guardduty.member.get(
        ctx,
        detector_id=detector_id,
        account_id=account_id,
        resource_id=resource_id,
        name=name,
    )
    if not before["result"]:
        result["result"] = before["result"]
        result["comment"] = before["comment"]
        return result
    if not before["ret"]:
        # create new resource
        update_needed = True
    else:
        # check if existing member needs to be added
        member_details = before["ret"]
        if member_details.get("relationship_status") not in ("Created", "Enabled"):
            update_needed = True
    if update_needed:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "detector_id": detector_id,
                    "account_id": account_id,
                    "email": email,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.guardduty.member", name=name
            )
            return result
        ret = await hub.exec.boto3.client.guardduty.create_members(
            ctx, DetectorId=detector_id, AccountDetails=revised_account_details
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.guardduty.member", name=name
        )
        resource_id = detector_id + ":" + account_id
    else:
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.guardduty.member", name=name
        )
    if update_needed:
        after = await hub.exec.aws.guardduty.member.get(
            ctx,
            detector_id=detector_id,
            account_id=account_id,
            resource_id=resource_id,
            name=name,
        )
        if not after["result"]:
            result["result"] = after["result"]
            result["comment"] += tuple(after["comment"])
            return result
        result["new_state"] = after["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    detector_id: str = None,
    account_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Delete member from AWS Guardduty.

    Deletes GuardDuty member account specified by the account ID.

    Args:
        name(str):
            An Idem name of the resource.

        detector_id(str, Optional):
            AWS Guardduty Detector id

        account_id(str, Optional):
            AWS Guardduty member account id

        resource_id(str, Optional):
            An identifier refers to an existing resource. The format is <detector_id>:<account_id>

    Request Syntax:
      .. code-block:: sls
        [idem_test_aws_guardduty_member]:
          aws.guardduty.member.absent:
            - name: 'string'
            - detector_id: 'string'
            - account_id: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.guardduty.member.absent:
                - name: test_delete_members
                - detector_id: '68c25425ab84ea0dcae26311eddacd34'
                - account_id: '106828723025'
                - resource_id: '68c25425ab84ea0dcae26311eddacd34:106828723025'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.guardduty.member", name=name
        )
        return result

    if not detector_id or not account_id:
        result["comment"] = hub.tool.aws.comment_utils.missing_args_for_absent_comment(
            resource_type="aws.guardduty.member",
            name=name,
            args=["detector_id", "account_id"],
        )
        result["result"] = False
        return result

    before = await hub.exec.aws.guardduty.member.get(
        ctx,
        detector_id=detector_id,
        account_id=account_id,
        resource_id=resource_id,
        name=name,
    )

    if not before["result"]:
        result["result"] = before["result"]
        result["comment"] = before["comment"]
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.guardduty.member", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.guardduty.member", name=name
        )
        return result
    else:
        account_ids = [account_id]
        result["old_state"] = before["ret"]
        try:
            ret = await hub.exec.boto3.client.guardduty.delete_members(
                ctx, DetectorId=detector_id, AccountIds=account_ids
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.guardduty.member", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    List details about all member accounts for the current GuardDuty administrator account.
    Describe returns detector_id, master_id, relationship_status, invited_at, updated_at, administrator_id as additional params
    not used in present input

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.guardduty.member
    """

    result = {}
    ret = await hub.exec.boto3.client.guardduty.list_detectors(ctx)
    if not ret["ret"]["DetectorIds"]:
        hub.log.debug(f"Could not list detector {ret['comment']}")
        return {}

    for detector_id in ret["ret"]["DetectorIds"]:
        members = await hub.exec.aws.guardduty.member.list(
            ctx,
            detector_id=detector_id,
            name=f"Describe members for detector {detector_id}",
        )
        if not members["result"]:
            hub.log.debug(f"Could not find members for detector {detector_id}")
            continue

        for member in members["ret"]:
            account_id = member.get("account_id")
            resource_id = detector_id + ":" + account_id
            # to avoid any other state overriding describe if just resource_id is the state_id
            resource_key = f"aws-guardduty-member-{resource_id}"
            result[resource_key] = {
                "aws.guardduty.member.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in member.items()
                ]
            }
    return result



from typing import Any
from typing import Dict
from typing import List


async def is_resource_updated(
    hub,
    before: Dict[str, Any],
    role_arn: str,
    recording_group: Dict,
):
    if role_arn != before.get("role_arn"):
        return True

    if recording_group != before.get("recording_group"):
        return True

    return False


async def is_account_agg_source_list_equal(
    hub, new_list: List[Dict], old_list: List[Dict]
):
    new_list.sort()
    old_list.sort()
    return new_list == old_list


async def is_configuration_aggregator_updated(
    hub,
    before: Dict[str, Any],
    account_aggregation_sources: List[Dict],
    organization_aggregation_source: Dict,
):
    # Flag to check if resource is updated or not
    if (
        account_aggregation_sources is not None
        and before.get("account_aggregation_sources") is not None
    ):
        if not await hub.tool.aws.config.config_utils.is_account_agg_source_list_equal(
            account_aggregation_sources, before.get("account_aggregation_sources")
        ):
            return True
    if (
        account_aggregation_sources is None
        and before.get("account_aggregation_sources") is not None
    ) or (
        account_aggregation_sources is not None
        and before.get("account_aggregation_sources") is None
    ):
        return True
    return not hub.tool.aws.state_comparison_utils.compare_dicts(
        organization_aggregation_source, before.get("organization_aggregation_source")
    )



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

from dict_tools import data


async def update_events_rule(
    hub,
    ctx,
    name,
    plan_state: Dict[str, Any],
    resource_id: str,
    old_targets: List = None,
    new_targets: List = None,
    schedule_expression: str = None,
    event_pattern: str = None,
    state: str = None,
    description: str = None,
    role_arn: str = None,
    event_bus_name: str = None,
):
    """Updates the specified rule.

    If you are updating an existing rule, the rule is replaced with what you specify in this
    update_events_rule. If you omit arguments in update_events_rule , the old values for those arguments are not kept.
    Instead, they are replaced with null values.

    When you create or update a rule, incoming events might not immediately start matching to new or updated rules.
    Allow a short period of time for changes to take effect.

    A rule must contain at least an event_pattern or schedule_expression. Rules with event_pattern are triggered when a
    matching event is observed. Rules with schedule_expression self-trigger based on the given schedule. A rule can
    have both an event_pattern and a schedule_expression, in which case the rule triggers on matching events as well as
    on a schedule.

    Args:
        plan_state(dict): idem ``--test`` state for update on AWS Lambda.
        resource_id: The name of the AWS CloudWatch Events Rule.
        old_targets(list, Optional): Existing targets associated with given CloudWatchEvent Rule.
        new_targets(list, Optional): The targets to update or add to the rule.
        schedule_expression (str, Optional): Scheduling expression.
            For example, ``"cron(0 20 * * ? *)"`` or ``"rate(5 minutes)"``.
        event_pattern (str, Optional): Rules use event patterns to select events and route them to targets. A pattern
            either matches an event or it doesn't. Event patterns are represented as JSON objects with a structure that
            is similar to that of events.
        state (str, Optional): Indicates whether the rule is enabled or disabled.
        description (str, Optional): A description of the rule.
        role_arn (str, Optional): The Amazon Resource Name (ARN) of the IAM role associated with the rule.
            If you're setting an event bus in another account as the target and that account granted permission to your
            account through an organization instead of directly by the account ID, you must specify a RoleArn with proper
            permissions in the Target structure, instead of here in this parameter.
        event_bus_name (str, Optional) : The name or ARN of the event bus to

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    update_params = OrderedDict(
        {
            "rule_status": state,
            "schedule_expression": schedule_expression,
            "event_bus_name": event_bus_name,
            "description": description,
            "role_arn": role_arn,
            "name": resource_id,
        }
    )
    modified_params = {}
    for key, value in update_params.items():
        if value and value != plan_state.get(key):
            modified_params[key] = value
            plan_state[key] = value

    if event_pattern and not hub.tool.aws.state_comparison_utils.are_policies_equal(
        plan_state.get("event_pattern", "{}"), event_pattern
    ):
        modified_params["event_pattern"] = event_pattern
        plan_state["event_pattern"] = event_pattern

    if modified_params:
        if not ctx.get("test", False):
            update_rule = await hub.exec.boto3.client.events.put_rule(
                ctx=ctx,
                Name=plan_state.get("name"),
                ScheduleExpression=plan_state.get("schedule_expression"),
                EventPattern=plan_state.get("event_pattern"),
                State=plan_state.get("rule_status"),
                Description=plan_state.get("description"),
                RoleArn=plan_state.get("role_arn"),
                EventBusName=plan_state.get("event_bus_name"),
            )
            result["result"] = update_rule["result"]
            if not result["result"]:
                result["comment"] = update_rule["comment"]
                return result
    target_diff = data.recursive_diff(old_targets, new_targets, ignore_order=True)
    if new_targets and target_diff:
        modified_params["targets"] = new_targets
        plan_state["targets"] = new_targets
        targets_to_put = target_diff.get("new")
        targets_to_remove = target_diff.get("old")
        if not ctx.get("test", False):
            if targets_to_remove:
                remove_target = await hub.exec.boto3.client.events.remove_targets(
                    ctx,
                    Rule=plan_state.get("name"),
                    Ids=[target["Id"] for target in targets_to_remove],
                    EventBusName=plan_state.get("event_bus_name"),
                )
                result["result"] = remove_target["result"]
                if not result["result"]:
                    result["comment"] += remove_target["comment"]
                    return result
            if targets_to_put:
                put_targets = await hub.exec.boto3.client.events.put_targets(
                    ctx,
                    Rule=plan_state.get("name"),
                    Targets=targets_to_put,
                    EventBusName=plan_state.get("event_bus_name"),
                )
                result["result"] = put_targets["result"]
                if not result["result"]:
                    result["comment"] += put_targets["comment"]
                    return result
    if modified_params:
        if not ctx.get("test", False):
            result["comment"] += hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.events.rule", name=name
            )
        else:
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.events.rule", name=name
            )
        result["ret"] = modified_params
    return result


def compare_and_return_target_ids(old_targets: List, new_targets: List):
    """Compares old_targets and new targets and return the new list of target ids that need to be updated.

    Args:
        old_targets(list): Existing targets associated with the given rule.
        new_targets(list): Newer list of targets to be associated with rule.

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    targets_to_remove = []
    old_targets_map = {target.get("Id"): target for target in old_targets or []}
    if new_targets is not None:
        for target in new_targets:
            if target.get("Id") in old_targets_map:
                if target != old_targets_map.get(target.get("Id")):
                    targets_to_remove.append(target.get("Id"))

        old_targets_map = {target.get("Id"): target for target in old_targets or []}
        for id in targets_to_remove:
            del old_targets_map[id]
        final_targets = list(old_targets_map.values()) + new_targets
        result["ret"] = {"final_result": final_targets}
    return result



"""Exec module for managing Amazon DocDB DB Cluster."""
from typing import List

__func_alias__ = {"list_": "list"}

RESOURCE_TYPE = "aws.docdb.db_cluster"


async def get(hub, ctx, name: str, resource_id: str = None):
    """Retrieves the specified AWS Document DB Cluster.

    Args:
        resource_id(str, Optional):
            AWS Document DB Cluster Identifier. If resource_id is not given name will be used to retrieve the DB Cluster.
        name(str):
            An Idem name of the resource.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.docdb.db_cluster.get resource_id="resource_id" name="name"

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.docdb.db_cluster.get(
                    ctx, name=name, resource_id=resource_id
                )

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.docdb.db_cluster.get
                - kwargs:
                    name: name
                    resource_id: resource_id

    """
    result = dict(comment=[], result=True, ret=None, resource_arn=None)

    raw_db_cluster_result = await hub.exec.boto3.client.docdb.describe_db_clusters(
        ctx, DBClusterIdentifier=resource_id if resource_id else name
    )
    if not raw_db_cluster_result["result"]:
        if "DBClusterNotFoundFault" in str(raw_db_cluster_result["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type=RESOURCE_TYPE,
                    name=name,
                )
            )
            result["comment"].append(raw_db_cluster_result["comment"])
            return result
        result["result"] = False
        result["comment"].append(raw_db_cluster_result["comment"])
        return result

    # if we don't find db_cluster with provided resource id, fail the request
    if not (raw_db_cluster_result["ret"].get("DBClusters")):
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type=RESOURCE_TYPE,
                name=name,
            )
        )
        return result
    if len(raw_db_cluster_result["ret"].get("DBClusters")) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_id=resource_id, resource_type=RESOURCE_TYPE
            )
        )
    resource_arn = raw_db_cluster_result["ret"].get("DBClusters")[0].get("DBClusterArn")
    tags = await hub.tool.aws.docdb.tag.get_tags_for_resource(
        ctx, resource_arn=resource_arn
    )
    # if failure while fetching tags, fail the request
    if not tags["result"]:
        result["result"] = False
        result["comment"].append(tags["comment"])
        return result
    tags = tags["ret"]
    result["ret"] = hub.tool.aws.docdb.db_cluster.convert_raw_db_cluster_to_present(
        raw_resource=raw_db_cluster_result["ret"]["DBClusters"][0],
        idem_resource_name=name,
        tags=tags,
    )
    return result


async def list_(hub, ctx, name: str = None, filters: List = None):
    """

    Args:
        name(str, Optional):
            The name of idem state.
        filters(list, Optional):
            One or more filters, A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/docdb.html#DocDB.Client.describe_db_clusters

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS Doc DB Cluster in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.docdb.db_cluster.list name="idem_name"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.docdb.db_cluster.list
                - kwargs:
                    name: dbg1
                    filters:
                      - name: filter-name
                        values:
                            - filter_value


    """
    result = dict(comment=[], ret=[], result=True)

    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result

    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )

    ret = await hub.exec.boto3.client.docdb.describe_db_clusters(
        ctx, Filters=boto3_filter
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["DBClusters"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        )
        return result
    result["comment"] = list(ret["comment"])
    for db_cluster in ret["ret"]["DBClusters"]:
        db_cluster_identifier = db_cluster.get("DBClusterIdentifier")
        arn = db_cluster["DBClusterArn"]
        tags = await hub.tool.aws.docdb.tag.get_tags_for_resource(ctx, resource_arn=arn)
        if not tags["result"]:
            result["result"] = False
            result["comment"].append(tags["comment"])
            return result
        tags = tags["ret"]
        result["ret"].append(
            hub.tool.aws.docdb.db_cluster.convert_raw_db_cluster_to_present(
                raw_resource=db_cluster,
                idem_resource_name=db_cluster_identifier,
                tags=tags,
            )
        )
    return result



"""State module for enabling organization admin account for Guardduty."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""Enables an Amazon Web Services account within the organization as the GuardDuty delegated administrator.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            An identifier of the resource in the provider. Here it is the Amazon Web Services Account ID for the organization account to be enabled as a GuardDuty
            delegated administrator.


    Request Syntax:
      .. code-block:: sls
            resource_is_present:
              aws.guardduty.organization_admin_account.present:
                - name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.guardduty.organization_admin_account.present:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_created = False
    before = await hub.exec.aws.guardduty.organization_admin_account.list(
        ctx, name=name
    )
    if not before["result"]:
        result["result"] = before["result"]
        result["comment"] = before["comment"]
        return result
    if before and before["ret"]:
        admin_accounts = before["ret"]
        # list call might return more than one account, so check whether it contains resource_id
        filtered_admin_accounts = filter(
            lambda admin: admin.get("resource_id") == resource_id
            and admin.get("admin_status") == "ENABLED",
            admin_accounts,
        )
        accounts = list(filtered_admin_accounts)
        if len(accounts) != 0:
            # resource_id is already an admin
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.guardduty.organization_admin_account", name=name
            )
            result["old_state"] = accounts[0]
        else:
            # some other account is an admin, so end the operation
            enabled_account_id = admin_accounts[0]["resource_id"]
            result["comment"] = (
                f"{enabled_account_id} is already enabled as organization_admin_account. "
                f"{resource_id} can only be enabled if {enabled_account_id} is removed as organization_admin_account"
            )
            result["result"] = False
            return result
    else:
        resource_created = True
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.guardduty.organization_admin_account", name=name
            )
            return result
        ret = await hub.exec.boto3.client.guardduty.enable_organization_admin_account(
            ctx, AdminAccountId=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.guardduty.organization_admin_account", name=name
        )
    if resource_created:
        # fetch resource
        after = await hub.exec.aws.guardduty.organization_admin_account.list(
            ctx, name=name
        )
        if not after["result"]:
            result["result"] = after["result"]
            result["comment"] = after["comment"]
            return result
        admin_accounts = after["ret"]
        filtered_admin_accounts = filter(
            lambda admin: admin.get("resource_id") == resource_id
            and admin.get("admin_status") == "ENABLED",
            admin_accounts,
        )
        account = list(filtered_admin_accounts)[0]
        result["new_state"] = account
    else:
        # already exists
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Disables an Amazon Web Services account within the Organization as the GuardDuty delegated administrator.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            The Amazon Web Services Account ID for the organizations account to be disabled as a GuardDuty
            delegated administrator.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.guardduty.organization_admin_account.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.aws.guardduty.organization_admin_account.list(
        ctx, name=name
    )
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.guardduty.organization_admin_account", name=name
        )
    else:
        admin_accounts = before["ret"]
        # list call might return more than one account, so check whether it contains resource_id
        filtered_admin_accounts = filter(
            lambda admin: admin.get("resource_id") == resource_id
            and admin.get("admin_status") == "ENABLED",
            admin_accounts,
        )
        eligible_accounts = list(filtered_admin_accounts)
        if len(eligible_accounts) != 0:
            # resource_id is the admin, disable it
            eligible_account = eligible_accounts[0]
            if ctx.get("test", False):
                result["old_state"] = eligible_account
                result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                    resource_type="aws.guardduty.organization_admin_account", name=name
                )
                return result
            else:
                result["old_state"] = eligible_account
                try:
                    ret = await hub.exec.boto3.client.guardduty.disable_organization_admin_account(
                        ctx, AdminAccountId=resource_id
                    )
                    result["result"] = ret["result"]
                    if not result["result"]:
                        result["comment"] = ret["comment"]
                        return result
                    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                        resource_type="aws.guardduty.organization_admin_account",
                        name=name,
                    )
                except hub.tool.boto3.exception.ClientError as e:
                    result["comment"] = result["comment"] + (
                        f"{e.__class__.__name__}: {e}",
                    )
        else:
            # already absent
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.guardduty.organization_admin_account", name=name
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists the accounts configured as GuardDuty delegated administrators.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.guardduty.organization_admin_account
    """
    result = {}
    ret = await hub.exec.aws.guardduty.organization_admin_account.list(
        ctx, name="list organization admins"
    )
    if not ret["result"]:
        hub.log.warning(f"Could not list organization_admin_account {ret['comment']}")
        return {}
    for organization_admin_account in ret["ret"]:
        resource_id = organization_admin_account["resource_id"]
        # to avoid any other state overriding describe if just resource_id is the state_id
        resource_key = f"aws-guardduty-org_admin_account-{resource_id}"
        result[resource_key] = {
            "aws.guardduty.organization_admin_account.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in organization_admin_account.items()
            ]
        }
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS CloudFront to present input format.
"""


def convert_raw_distribution_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    resource_id = raw_resource.get("Id")
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "arn": raw_resource.get("ARN"),
    }
    resource_parameters = OrderedDict(
        {
            "CallerReference": "caller_reference",
            "Comment": "comment",
            "Origins": "origins",
            "DefaultCacheBehavior": "default_cache_behaviour",
            "Logging": "logging",
            "Enabled": "enabled",
            "ViewerCertificate": "viewer_certificate",
            "Aliases": "aliases",
            "PriceClass": "price_class",
            "DefaultRootObject": "default_root_object",
            "OriginGroups": "origin_groups",
            "CacheBehaviors": "cache_behaviors",
            "CustomErrorResponses": "custom_error_responses",
            "Restrictions": "restrictions",
            "WebACLId": "web_acl_id",
            "HttpVersion": "http_version",
            "IsIPV6Enabled": "is_ipv6_enabled",
        }
    )

    distribution_config = raw_resource["DistributionConfig"]
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in distribution_config:
            resource_translated[parameter_present] = (
                dict(distribution_config.get(parameter_raw).copy())
                if isinstance(distribution_config.get(parameter_raw), dict)
                else distribution_config.get(parameter_raw)
            )

    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    if raw_resource.get("ETag") is not None:
        resource_translated["ETag"] = raw_resource.get("ETag")

    return hub.tool.aws.ec2.instance.data.sanitize_dict(resource_translated)


def convert_raw_cache_policy_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    resource_id = raw_resource.get("Id")
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
    }
    resource_parameters = OrderedDict(
        {
            "Comment": "comment",
            "Name": "name",
            "DefaultTTL": "default_ttl",
            "MaxTTL": "max_ttl",
            "MinTTL": "min_ttl",
            "ParametersInCacheKeyAndForwardedToOrigin": "parameter_in_cache_key_and_forward_to_origin",
        }
    )

    cache_policy_config = raw_resource["CachePolicyConfig"]
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in cache_policy_config:
            resource_translated[parameter_present] = (
                copy.deepcopy(dict(cache_policy_config.get(parameter_raw)))
                if isinstance(cache_policy_config.get(parameter_raw), dict)
                else cache_policy_config.get(parameter_raw)
            )
    if raw_resource.get("ETag") is not None:
        resource_translated["etag"] = raw_resource.get("ETag")
    resource_translated = hub.tool.aws.cloudfront.cloudfront_utils.sanitize_dict(
        resource_translated
    )
    return resource_translated


def convert_raw_origin_request_policy_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    resource_id = raw_resource.get("Id")
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
    }
    resource_parameters = OrderedDict(
        {
            "Comment": "comment",
            "Name": "name",
            "HeadersConfig": "headers_config",
            "CookiesConfig": "cookies_config",
            "QueryStringsConfig": "query_strings_config",
        }
    )

    origin_request_config = raw_resource["OriginRequestPolicyConfig"]
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in origin_request_config:
            resource_translated[parameter_present] = (
                dict(origin_request_config.get(parameter_raw).copy())
                if isinstance(origin_request_config.get(parameter_raw), dict)
                else origin_request_config.get(parameter_raw)
            )

    if raw_resource.get("ETag") is not None:
        resource_translated["etag"] = raw_resource.get("ETag")

    resource_translated = hub.tool.aws.cloudfront.cloudfront_utils.sanitize_dict(
        resource_translated
    )
    return resource_translated



"""Exec functions for s3 bucket's website configurations."""
from typing import Dict


async def get(
    hub, ctx, resource_id: str, name: str = None, expected_bucket_owner: str = None
) -> Dict:
    """Returns the website configuration of a bucket.

    Args:

        resource_id(str):
            AWS S3 bucket name.

        name(str, Optional):
            The name of the Idem state.

        expected_bucket_owner(str, Optional):
            The account ID of the expected bucket owner. If the bucket is owned by a different account, the
            request fails with the HTTP status code 403 Forbidden (access denied). Defaults to None.

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            $ idem exec aws.s3.bucket_website.get name="bucket-name" resource_id="bucket-id"

        Using in a state:

        .. code-block:: yaml

            get_a_bucket:
              exec.run:
                - path: aws.s3.bucket.get
                - kwargs:
                   name: bucket-name
                   resource_id: bucket-id
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.s3.get_bucket_website(
        ctx, Bucket=resource_id, ExpectedBucketOwner=expected_bucket_owner
    )
    if not ret["result"]:
        if "NoSuchWebsiteConfiguration" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.s3.bucket_website", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result[
        "ret"
    ] = hub.tool.aws.s3.bucket_website.convert_raw_bucket_website_to_present(
        bucket=resource_id, raw_resource=ret["ret"], idem_resource_name=name
    )
    return result



"""Utility functions for Rds Db Proxy Targets."""
from typing import Any
from typing import Dict


async def convert_raw_resource_to_present_async(
    hub,
    ctx,
    idem_resource_name: str,
    resource_id: str,
    raw_resource: dict,
    db_proxy_name: str,
    target_group_name: str,
) -> Dict[str, Any]:
    r"""
    Convert raw resource of db_proxy_target type into present format.

    """

    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "db_proxy_name": db_proxy_name,
        "target_group_name": target_group_name,
    }

    resource_parameters = {
        "TargetArn": "target_arn",
        "Endpoint": "endpoint",
        "Port": "port",
        "Type": "type",
        "Role": "role",
        "TargetHealth": "target_health",
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if "TrackedClusterId" in raw_resource:
        resource_translated["db_cluster_identifiers"] = [
            raw_resource.get("TrackedClusterId")
        ]
    if "RdsResourceId" in raw_resource:
        resource_translated["db_instance_identifiers"] = [
            raw_resource.get("RdsResourceId")
        ]

    return resource_translated


def create_resource_id(hub, raw_resource, db_proxy_name, target_group_name):
    return (
        db_proxy_name
        + "/"
        + target_group_name
        + "/"
        + raw_resource.get("Type")
        + "/"
        + raw_resource.get("RdsResourceId")
    )



"""State module for managing Amazon Organizations Accounts."""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
SERVICE = "organizations"


TREQ = {
    "absent": {
        "require": [
            "aws.organizations.policy_attachment.absent",
            "aws.organizations.policy.absent",
        ],
    },
    "present": {
        "require": [
            "aws.organizations.organization_unit.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    email: str,
    role_name: str = None,
    iam_user_access_to_billing: str = "ALLOW",
    resource_id: str = None,
    parent_id: str = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
) -> Dict[str, Any]:
    """Creates an AWS account that is automatically a member of the organization whose credentials made the request.

    Args:
        name(str):
            The name of the member account.
        email(str):
            The email address of the owner to assign to the new member account.
            This email address must not already be associated with another Amazon Web Services account.
            You must use a valid email address to complete account creation.
        role_name(str, Optional):
            The name of an IAM role that Organizations automatically preconfigures in the new member account.
            This role trusts the management account, allowing users in the management account to assume the role,
            as permitted by the management account administrator. The role has administrator permissions in the new member account.
            If you don't specify this parameter, the role name defaults to ``OrganizationAccountAccessRole``.
        iam_user_access_to_billing(str, Optional):
            If set to ``ALLOW``, the new account enables IAM users to access account billing information if they have the required permissions.
            If set to ``DENY``, only the root user of the new account can access account billing information.
            If you don't specify this parameter, the value defaults to ``ALLOW``.
        resource_id(str, Optional):
            The ID of the member account in Amazon Web Services.
        parent_id(str, Optional):
            Parent Organizational Unit ID or Root ID for the account. Defaults to the Organization default Root ID.
        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the member account.

            * Key (*str*):
                The key identifier, or name, of the tag.
            * Value (*str*):
                The string value that's associated with the key of the tag.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_organizations_account]:
          aws.organizations.account.present:
            - name: 'string'
            - resource_id: 'string'
            - email: 'string'
            - role_name: 'string'
            - iam_user_access_to_billing: 'ALLOW|DENY'
            - parent_id: 'string'
            - tags:
              - Key: 'string'
                Value: 'string

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_account:
              aws.organizations.account.present:
                - name: 'idem_test_account'
                - email: 'xyz@email.com'
                - role_name: 'idem_test_role'
                - iam_user_access_to_billing: 'ALLOW'
                - parent_id: 'o-parent-id'
                - tags:
                  - Key: 'provider'
                    Value: 'idem'
    """
    result = dict(comment=(), name=name, result=True, old_state=None, new_state=None)
    before = None
    update_tag = False
    update_parent = False

    if resource_id:
        before = await hub.exec.aws.organizations.account.get(
            ctx, resource_id=resource_id, name=name
        )
        if not before["result"]:
            result["result"] = before["result"]
            result["comment"] = before["comment"]
            return result

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before and before["ret"]:
        # Account exists , update
        try:
            result["old_state"] = before["ret"]
            plan_state = copy.deepcopy(result["old_state"])
            # we need to list parents to check if move is required in case the new_parent_id is not equal to current_
            # parent_id
            current_parent_id = before["ret"]["parent_id"]
            if current_parent_id and parent_id and current_parent_id != parent_id:
                if not ctx.get("test", False):
                    move_account_result = (
                        await hub.tool.aws.organizations.account.move_account(
                            ctx, resource_id, result, parent_id
                        )
                    )
                    if move_account_result and not move_account_result["result"]:
                        result["comment"] = result["comment"] + (
                            f"Could not update parent for aws.organizations.account {name}",
                        )
                        result["result"] = False
                        return result
                update_parent = True
            if update_parent:
                plan_state["parent_id"] = parent_id
                result["comment"] = result["comment"] + (
                    f"Would update parent for aws.organizations.account {name}",
                )

            # update tags
            if tags is not None and tags != result["old_state"].get("tags"):
                update_tags_ret = await hub.tool.aws.organizations.tag.update_tags(
                    ctx, resource_id, result["old_state"].get("tags"), tags
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = update_tags_ret["result"]
                    return result

                result["comment"] = result["comment"] + (
                    f"Updated tags on aws.organizations.account '{name}'.",
                )
                update_tag = True
                if ctx.get("test", False):
                    if update_tags_ret["ret"]:
                        plan_state["tags"] = update_tags_ret["ret"]
                    else:
                        plan_state["tags"] = result["old_state"]["tags"]

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    else:
        # Account not present , create
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "email": email,
                    "name": name,
                    "parent_id": parent_id,
                    "role_name": role_name,
                    "iam_user_access_to_billing": iam_user_access_to_billing,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.organizations.account", name=name
            )
            return result

        try:
            create_account_ret = (
                await hub.exec.boto3.client.organizations.create_account(
                    ctx,
                    Email=email,
                    AccountName=name,
                    RoleName=role_name,
                    IamUserAccessToBilling=iam_user_access_to_billing,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                )
            )

            result["result"] = create_account_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + create_account_ret["comment"]
                return result

            account_status_id = create_account_ret["ret"]["CreateAccountStatus"]["Id"]

            # Call a custom waiter to wait on account's creation.

            acceptors = [
                {
                    "matcher": "path",
                    "expected": "SUCCEEDED",
                    "state": "success",
                    "argument": "CreateAccountStatus.State",
                },
                {
                    "matcher": "path",
                    "expected": "IN_PROGRESS",
                    "state": "retry",
                    "argument": "CreateAccountStatus.State",
                },
                {
                    "matcher": "path",
                    "expected": "FAILED",
                    # Failure is also mapped with success to catch the error message
                    "state": "success",
                    "argument": "CreateAccountStatus.State",
                },
            ]
            account_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                name="AccountCreated",
                operation="DescribeCreateAccountStatus",
                argument=["CreateAccountStatus.State"],
                acceptors=acceptors,
                client=await hub.tool.boto3.client.get_client(ctx, SERVICE),
                matcher="path",
                delay=10,
                max_tries=10,
            )
            await hub.tool.boto3.client.wait(
                ctx,
                SERVICE,
                "AccountCreated",
                account_waiter,
                CreateAccountRequestId=account_status_id,
            )

            account_status = await hub.exec.boto3.client.organizations.describe_create_account_status(
                ctx, CreateAccountRequestId=account_status_id
            )
            if account_status["result"]:
                create_account_status = account_status["ret"]["CreateAccountStatus"]
                account_state = create_account_status["State"]
                if account_state == "FAILED":
                    result["result"] = False
                    result["comment"] = result["comment"] + (
                        create_account_status["FailureReason"],
                    )
                    return result
                elif account_state == "SUCCEEDED":
                    resource_id = create_account_status["AccountId"]
                    result["comment"] = hub.tool.aws.comment_utils.create_comment(
                        resource_type="aws.organizations.account", name=name
                    )
                    if resource_id is not None and parent_id is not None:
                        parents = (
                            await hub.exec.boto3.client.organizations.list_parents(
                                ctx, ChildId=resource_id
                            )
                        )
                        result["result"] = result["result"] and parents["result"]
                        if not result["result"]:
                            result["comment"] = result["comment"] + parents["comment"]
                            return result
                        if parents:
                            current_parent_id = parents["ret"]["Parents"][0]["Id"]
                            if current_parent_id != parent_id:
                                move_account_result = await hub.tool.aws.organizations.account.move_account(
                                    ctx, resource_id, result, parent_id
                                )
                                if (
                                    move_account_result
                                    and not move_account_result["result"]
                                ):
                                    result["comment"] = result["comment"] + (
                                        f"Could not update parent for aws.organizations.account {name}",
                                    )
                                    result["result"] = False
                                    return result
            else:
                result["result"] = False
                result["comment"] = result["comment"] + account_status["comment"]
                return result

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not before or update_parent or update_tag and resource_id:
        try:
            after = await hub.exec.aws.organizations.account.get(
                ctx, resource_id=resource_id, name=name
            )
            if after and after.get("ret"):
                result["new_state"] = after["ret"]
            else:
                result["result"] = result["result"] and after["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + after["comment"]
                    return result

        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Removes the specified account from the organization.

    The removed account becomes a standalone account that isn't a member of any organization.
    It's no longer subject to any policies and is responsible for its own bill payments.
    The organization's management account is no longer charged for any expenses accrued by
    the member account after it's removed from the organization.
    This operation can be called only from the organization's management account.

    Args:
        name(str):
            The name of the member account.
        resource_id(str, Optional):
            The ID of the member account in Amazon Web Services.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_organizations_account]:
          aws.organizations.account.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_account:
              aws.organizations.account.absent:
                - name: 'idem_test_account'
                - resource_id: '123456789012'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.organizations.account",
            name=name,
        )
        return result

    before = await hub.exec.aws.organizations.account.get(
        ctx, resource_id=resource_id, name=name
    )

    if not before["result"]:
        result["result"] = before["result"]
        result["comment"] = before["comment"]
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.organizations.account",
            name=name,
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.organizations.account",
            name=name,
        )
    else:
        try:
            ret = await hub.exec.boto3.client.organizations.remove_account_from_organization(
                ctx, AccountId=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.organizations.account", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if before:
        result["old_state"] = before["ret"]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes AWS Organizations Accounts in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.organizations.account
    """
    result = {}

    describe_ret = await hub.exec.aws.organizations.account.list(ctx)
    if not describe_ret["ret"]:
        hub.log.debug(f"Could not list accounts {describe_ret['comment']}")
        return {}

    accounts = describe_ret["ret"]

    for account in accounts:
        account_id = account["resource_id"]
        result[account_id] = {
            "aws.organizations.account.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in account.items()
            ]
        }
    return result



from typing import Any
from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: List[Dict[str, Any]],
    new_value: List[Dict[str, Any]],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in its "present" state

    - network_interfaces:
        - network_interface_id: 'string'
          device_index: 123
          network_card_index: 123

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    result = True
    # attach/detach network interfaces as needed
    new_interfaces = {n.get("network_interface_id"): n for n in new_value}
    old_interfaces = {n.get("network_interface_id"): n for n in old_value}

    interfaces_to_attach = set(new_interfaces.keys()) - set(old_interfaces.keys())
    interfaces_to_delete = set(old_interfaces.keys()) - set(new_interfaces.keys())
    for network_interface_id, old_interface in old_interfaces.items():
        new_interface = new_interfaces.get(network_interface_id, {})
        if not new_interface:
            comments.append(
                f"Network Interface removed from instance: {network_interface_id}"
            )
            # Remove the interface
            interfaces_to_delete.add(network_interface_id)
        else:
            if old_interface.get("device_index") != new_interface.get(
                "device_index"
            ) and new_interface.get("device_index"):
                comments.append(
                    "Network Interface moved from device index "
                    f"{old_interface.get('device_index')} to {new_interface.get('device_index')}"
                )
                # Remove the interface and attach it to the new index
                interfaces_to_delete.add(network_interface_id)
                interfaces_to_attach.add(network_interface_id)
            if old_interface.get("network_card_index") != new_interface.get(
                "network_card_index"
            ) and new_interface.get("network_card_index"):
                comments.append(
                    "Network Interface moved from network card index "
                    f"{old_interface.get('network_card_index')} to {new_interface.get('network_card_index')}"
                )
                # Remove the interface and attach it to the new index
                interfaces_to_delete.add(network_interface_id)
                interfaces_to_attach.add(network_interface_id)

    # detach interfaces
    for network_interface_id in interfaces_to_delete:
        attachment_ret = (
            await hub.exec.boto3.client.ec2.describe_network_interface_attribtue(
                ctx, Attribute="attachment", NetworkInterfaceId=network_interface_id
            )
        )
        if not attachment_ret.result:
            comments.append(attachment_ret.comment)
            continue

        attachment = attachment_ret.ret.get("Attachment", {})

        # detach the network interface from the instance
        ret = await hub.exec.boto3.client.ec2.attach_network_interface(
            AttachmentId=attachment.get("AttachmentId")
        )
        result &= ret.result
        if not ret.result:
            comments.append(ret.comment)

    # attach interfaces
    for network_interface_id in interfaces_to_attach:
        network_interface = new_interfaces[network_interface_id]
        kwargs = {}
        if network_interface.get("device_index") is not None:
            kwargs["DeviceIndex"] = network_interface["device_index"]
        if network_interface.get("network_card_index") is not None:
            kwargs["NetworkCardIndex"] = network_interface["network_card_index"]
        ret = await hub.exec.boto3.client.ec2.attach_network_interface(
            ctx,
            NetworkInterfaceId=network_interface_id,
            InstanceId=resource.id,
            **kwargs,
        )
        result &= ret.result
        if not ret.result:
            comments.append(ret.comment)

    return result



from typing import Any
from typing import Dict


def convert_to_present(
    hub, describe_network_interfaces: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Convert the output of describe_network_interfaces to present states format
    """
    result = {}

    for network_interface in describe_network_interfaces.get("NetworkInterfaces", ()):
        resource_id = network_interface.get("NetworkInterfaceId")

        association = network_interface.get("Association", {})
        attachment = network_interface.get("Attachment", {})
        ena_srd_specification = attachment.get("ena_srd_specification", {})

        result[resource_id] = dict(
            name=resource_id,
            resource_id=resource_id,
            # Association
            allocation_id=association.get("AllocationId"),
            ip_owner_id=association.get("IpOwnerId"),
            public_dns_name=association.get("PublicDnsName"),
            public_ip=association.get("PublicIp"),
            customer_owned_ip=association.get("CustomerOwnedIp"),
            carrier_ip=association.get("CarrierIp"),
            # Attachment
            attachment_id=attachment.get("AttachmentId"),
            delete_on_termination=attachment.get("DeleteOnTermination"),
            device_index=attachment.get("DeviceIndex"),
            network_card_index=attachment.get("NetworkCardIndex"),
            ena_srd_enabled=ena_srd_specification.get("EnaSrdEnabled"),
            ena_srd_udp_enabled=ena_srd_specification.get("EnaSrdUdpEnabled"),
            # General
            client_token=None,
            availability_zone=network_interface.get("AvailabilityZone"),
            description=network_interface.get("Description"),
            groups=[
                group.get("GroupId") for group in network_interface.get("Groups", [])
            ],
            interface_type=network_interface.get("InterfaceType"),
            ipv4_address_count=len(network_interface.get("Ipv4Addresses", [])) or None,
            ipv4_prefix_count=len(network_interface.get("Ipv4Prefixes", [])) or None,
            ipv4_prefixes=[
                i.get("Ipv4Prefix") for i in network_interface.get("Ipv4Prefixes", [])
            ]
            or None,
            ipv6_addresses=[
                i.get("Ipv6Address") for i in network_interface.get("Ipv6Addresses", [])
            ]
            or None,
            ipv6_address_count=len(network_interface.get("Ipv6Addresses", [])) or None,
            ipv6_prefix_count=len(network_interface.get("Ipv6Prefixes", [])) or None,
            ipv6_prefixes=[
                i.get("Ipv6Prefix") for i in network_interface.get("Ipv6Prefixes", [])
            ]
            or None,
            mac_address=network_interface.get("MacAddress"),
            outpost_arn=network_interface.get("OutpostArn"),
            owner_id=network_interface.get("OwnerId"),
            private_dns_name=network_interface.get("PrivateDnsName"),
            private_ip_addresses=[
                i.get("PrivateIpAddress")
                for i in network_interface.get("PrivateIpAddresses", [])
            ]
            or None,
            primary_ip_address=network_interface.get("PrivateIpAddress"),
            requester_id=network_interface.get("RequesterId"),
            requester_managed=network_interface.get("RequesterManaged"),
            source_dest_check=network_interface.get("SourceDestCheck"),
            subnet_id=network_interface.get("SubnetId"),
            status=network_interface.get("Status"),
            tags={
                i.get("Key"): i.get("Value")
                for i in network_interface.get("TagSet", [])
            },
            vpc_id=network_interface.get("VpcId"),
            deny_all_igw_traffic=network_interface.get("DenyAllIgwTraffic"),
            ipv6_native=network_interface.get("Ipv6Native"),
            default_ipv6_address=network_interface.get("Ipv6Address"),
        )

    return result



import datetime
from typing import Any
from typing import Dict
from typing import List


def sanitize_dict(hub, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Iterate over a present state and verify that all the options are serializable
    """
    result = {}

    for key, value in data.items():
        if value is None:
            continue
        result[key] = hub.tool.aws.cloudfront.cloudfront_utils.sanitize_value(value)

    return result


def sanitize_list(hub, data: List[Any]) -> List[Any]:
    """
    Iterate over a present state and verify that all the options are serializable
    """
    result = []

    for value in data:
        if value is None:
            continue
        result.append(hub.tool.aws.cloudfront.cloudfront_utils.sanitize_value(value))

    return result


def sanitize_value(hub, data: Any) -> Any:
    """
    Iterate over a present state and verify that all the options are serializable
    """

    if isinstance(data, Dict):
        return hub.tool.aws.cloudfront.cloudfront_utils.sanitize_dict(data)
    elif isinstance(data, datetime.datetime):
        return str(data)
    elif isinstance(data, List):
        return hub.tool.aws.cloudfront.cloudfront_utils.sanitize_list(data)
    else:
        return data


def sanitize_policy_config(hub, original_policy_config: Dict[str, Any]):
    updated_policy_config = {}

    for (
        policy_option_key,
        policy_option_value,
    ) in original_policy_config.items():
        if policy_option_value is not None:
            updated_policy_config[policy_option_key] = policy_option_value

    return updated_policy_config


async def update_cache_policy(
    hub,
    ctx,
    name: str,
    before: Dict[str, Any],
    resource_id: str,
    comment: str,
    default_ttl: int,
    max_ttl: int,
    min_ttl: int,
    parameter_in_cache_key_and_forward_to_origin: Dict[str, Any],
    if_match: str,
):
    result = dict(comment=(), result=True, ret=None)
    # cloudfront cache_policy requires all params to be sent for update call even though one parameter change.
    # we will check if at least one attribute is changed and send the update request if any of the attribute changes.

    update_allowed_params = [
        "comment",
        "name",
        "default_ttl",
        "max_ttl",
        "min_ttl",
        "parameter_in_cache_key_and_forward_to_origin",
    ]

    modified_params = {}
    for parameter in update_allowed_params:
        if (
            locals()[parameter] is not None
            and before.get(parameter) != locals()[parameter]
        ):
            modified_params[parameter] = locals()[parameter]

    if modified_params:
        result["ret"] = modified_params
        if not ctx.get("test", False):
            cache_policy_config = {
                "Comment": comment,
                "Name": name,
                "DefaultTTL": default_ttl,
                "MaxTTL": max_ttl,
                "MinTTL": min_ttl,
                "ParametersInCacheKeyAndForwardedToOrigin": parameter_in_cache_key_and_forward_to_origin,
            }
            # Even though some params are optional in create, most of them are mandatory (at least default values) in
            # update call. Refer
            # https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/cache_policy-overview-required-fields.html
            # we will sanitize the cache_policy config and fill in missing fields with default values.

            cache_policy_config = (
                hub.tool.aws.cloudfront.cloudfront_utils.sanitize_policy_config(
                    cache_policy_config
                )
            )

            ret = await hub.exec.boto3.client.cloudfront.update_cache_policy(
                ctx,
                CachePolicyConfig=cache_policy_config,
                Id=resource_id,
                IfMatch=if_match,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

    return result


async def update_origin_request_policy(
    hub,
    ctx,
    name: str,
    before: Dict[str, Any],
    resource_id: str,
    comment: str,
    headers_config: Dict[str, Any],
    cookies_config: Dict[str, Any],
    query_strings_config: Dict[str, Any],
    if_match: str,
):
    result = dict(comment=(), result=True, ret=None)
    # cloudfront origin request policy requires all params to be sent for update call even though one parameter change.
    # we will check if at least one attribute is changed and send the update request if any of the attribute changes.

    update_allowed_params = [
        "comment",
        "name",
        "headers_config",
        "cookies_config",
        "query_strings_config",
    ]

    modified_params = {}
    for parameter in update_allowed_params:
        if (
            locals()[parameter] is not None
            and before.get(parameter) != locals()[parameter]
        ):
            modified_params[parameter] = locals()[parameter]

    if modified_params:
        result["ret"] = modified_params
        if not ctx.get("test", False):
            origin_request_policy_config = {
                "Comment": comment,
                "Name": name,
                "HeadersConfig": headers_config,
                "CookiesConfig": cookies_config,
                "QueryStringsConfig": query_strings_config,
            }
            # Even though some params are optional in create, most of them are mandatory (at least default values) in
            # update call. Refer
            # https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/cache_policy-overview-required-fields.html
            # we will sanitize the origin_request config and fill in missing fields with default values.

            origin_request_policy_config = (
                hub.tool.aws.cloudfront.cloudfront_utils.sanitize_policy_config(
                    origin_request_policy_config
                )
            )

            ret = await hub.exec.boto3.client.cloudfront.update_origin_request_policy(
                ctx,
                OriginRequestPolicyConfig=origin_request_policy_config,
                Id=resource_id,
                IfMatch=if_match,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

    return result



"""State module for managing Amazon API Gateway Stages."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    deployment_id: str,
    resource_id: str = None,
    description: str = None,
    cache_cluster_enabled: bool = None,
    cache_cluster_size: str = None,
    variables: dict = None,
    documentation_version: str = None,
    canary_settings: make_dataclass(
        "canarySettings",
        [
            ("percentTraffic", float),
            ("stageVariableOverrides", Dict, field(default=None)),
            ("deploymentId", str, field(default=None)),
        ],
    ) = None,
    tracing_enabled: bool = None,
    tags: dict = None,
):
    """Creates a new API Gateway Stage or modifies an existing one.

    Args:
        name(str):
            An Idem name of the resource. Stage names can only contain alphanumeric characters, hyphens, and underscores.
            Maximum length is 128 characters.

        rest_api_id(str):
            The string identifier of the associated RestApi.

        deployment_id(str):
            The identifier of the Deployment resource for the Stage resource.

        resource_id(str, Optional):
            AWS API Gateway Stage ID, in the form of '<rest_api_id>-<name>'.

        description(str, Optional):
            The description of the Stage resource.

        cache_cluster_enabled(bool, Optional):
            Whether cache clustering is enabled for the stage.

        cache_cluster_size(str, Optional):
            The stage's cache cluster size.

        variables(dict, Optional):
            A map that defines the stage variables for the new Stage resource. Variable names can have alphanumeric and
            underscore characters, and the values must match [A-Za-z0-9-._~:/?#&=,]+ .

        documentation_version(str, Optional):
            The version of the associated API documentation.

        canary_settings(dict, Optional):
            The canary deployment settings of this stage.

            * percentTraffic (float) --
            The percent (0-100) of traffic diverted to a canary deployment.

            * deploymentId (string) --
            The ID of the canary deployment.

            * stageVariableOverrides (dict) --
            Stage variables overridden for a canary release deployment, including new stage variables introduced in the
            canary. These stage variables are represented as a string-to-string map between stage variable names and
            their values. Dict of string: string.

            * useStageCache (boolean) --
            A Boolean flag to indicate whether the canary deployment uses the stage cache or not.

        tracing_enabled(bool, Optional):
            Specifies whether active tracing with X-ray is enabled for the Stage.

        tags(dict, Optional):
            The key-value map of strings. The valid character set is [a-zA-Z+-=._:/]. The tag key can be up to 128
            characters and must not start with aws: . The tag value can be up to 256 characters.

    Request Syntax:
        [idem_test_aws_apigateway_stage]:
          aws.apigateway.stage.present:
            - name: 'string'
            - description: 'string'
            - rest_api_id: 'string'
            - deployment_id: 'string'
            - resource_id: 'string-string'
            - cache_cluster_enabled: bool
            - cache_cluster_size: 'string'
            - variables:
                - 'string' : 'string'
            - canary_settings:
                - percentTraffic: float
                - deploymentId: 'string'
                - useStageCache: bool
            - tracing_enabled: bool
            - document_version: 'string'
            - tags:
                - 'string': 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.apigateway.stage.present:
                - name: value
                - rest_api_id: value
                - deployment_id: value

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.apigateway.stage.get(
            ctx,
            resource_id=resource_id,
        )

        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = before["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])

        updateable_parameters = {
            "cache_cluster_enabled": cache_cluster_enabled,
            "cache_cluster_size": cache_cluster_size,
            "deployment_id": deployment_id,
            "description": description,
            "variables": variables,
            "tracing_enabled": tracing_enabled,
        }

        update_stage_ret = await hub.tool.aws.apigateway.stage.update(
            ctx,
            old_state=result["old_state"],
            updateable_parameters=updateable_parameters,
        )

        result["comment"] = result["comment"] + update_stage_ret["comment"]

        if not update_stage_ret["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_stage_ret["ret"])
        if resource_updated and ctx.get("test", False):
            result["new_state"] = update_stage_ret["ret"]

        resource_arn = hub.tool.aws.arn_utils.build(
            service="apigateway",
            region=ctx["acct"]["region_name"],
            resource="/restapis/" + rest_api_id + "/stages/" + name,
        )

        if tags is not None and tags != result["old_state"].get("tags", {}):
            update_tags_ret = await hub.exec.aws.apigateway.tag.update_tags(
                ctx,
                resource_arn=resource_arn,
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags,
            )
            if not update_tags_ret["result"]:
                result["result"] = False
                result["comment"] = update_tags_ret["comment"]
                return result
            result["comment"] = result["comment"] + update_tags_ret["comment"]

            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if update_tags_ret["ret"] and ctx.get("test", False):
                result["new_state"]["tags"] = update_tags_ret["ret"]

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "description": description,
                    "rest_api_id": rest_api_id,
                    "deployment_id": deployment_id,
                    "cache_cluster_enabled": cache_cluster_enabled,
                    "cache_cluster_size": cache_cluster_size,
                    "variables": variables,
                    "canary_settings": canary_settings,
                    "tracing_enabled": tracing_enabled,
                    "documentation_version": documentation_version,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.stage", name=name
            )
            return result

        resource_id = f"{rest_api_id}-{name}"

        ret = await hub.exec.boto3.client.apigateway.create_stage(
            ctx,
            restApiId=rest_api_id,
            stageName=name,
            deploymentId=deployment_id,
            description=description,
            cacheClusterEnabled=cache_cluster_enabled,
            cacheClusterSize=cache_cluster_size,
            variables=variables,
            documentationVersion=documentation_version,
            tracingEnabled=tracing_enabled,
            canarySettings=canary_settings,
            tags=tags,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.stage", name=name
        )

    if ctx.get("test", False) and resource_updated:
        return result

    elif (not before) or resource_updated:
        after = await hub.exec.aws.apigateway.stage.get(
            ctx,
            resource_id=resource_id,
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result

        result["new_state"] = after["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes a Stage resource.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            AWS API Gateway Stage ID. Idem automatically considers this resource being absent if this field is not
            specified. in the form of '<rest_api_id>-<name>'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.apigateway.stage.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        # if resource_id isn't specified, the resource is considered to be absent.
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.stage", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.stage.get(
        ctx,
        resource_id=resource_id,
    )

    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.stage", name=name
        )
        return result

    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.stage", name=name
        )
        return result

    else:
        rest_api_id, name = resource_id.split("-")
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigateway.delete_stage(
            ctx, restApiId=rest_api_id, stageName=name
        )

        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.stage", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the API Gateway Stages.

    Returns a list of apigateway.stage descriptions

    Returns:
        Dict[str, Any]


    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.stage

    """

    result = {}
    get_rest_apis_ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)

    if not get_rest_apis_ret["result"]:
        hub.log.debug(f"Could not get Rest Apis {get_rest_apis_ret['comment']}")
        return result

    if get_rest_apis_ret["ret"]["items"] is not None:
        for rest_api in get_rest_apis_ret["ret"]["items"]:
            rest_api_id = rest_api.get("id")
            get_deployments_ret = (
                await hub.exec.boto3.client.apigateway.get_deployments(
                    ctx, restApiId=rest_api_id
                )
            )
            if not get_deployments_ret["result"]:
                hub.log.debug(
                    f"Could not describe deployments for Rest API {get_deployments_ret['comment']}. Skipping this resource "
                    f"and continuing."
                )
                continue

            for deployment in get_deployments_ret["ret"]["items"]:
                resource_id = deployment.get("resource_id")
                get_stages = await hub.exec.boto3.client.apigateway.get_stages(
                    ctx,
                    restApiId=rest_api_id,
                    deploymentId=resource_id,
                )
                if not get_stages["result"]:
                    hub.log.debug(f"Could not get Stages {get_stages['comment']}")
                    continue

                for stage in get_stages["ret"]["item"]:
                    stage_name = stage.get("stageName")
                    stage_id = f"{rest_api_id}-{stage_name}"
                    resource_translated = (
                        hub.tool.aws.apigateway.stage.convert_raw_stage_to_present(
                            raw_resource=stage,
                            resource_id=stage_id,
                        )
                    )
                    result[resource_translated["resource_id"]] = {
                        "aws.apigateway.stage.present": [
                            {parameter_key: parameter_value}
                            for parameter_key, parameter_value in resource_translated.items()
                        ]
                    }

    return result
